﻿using Labs.Utility;
using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;
using System;

namespace Labs.ACW
{
    public class ACWWindow : GameWindow
    {
        public ACWWindow()
            : base(
                800, // Width
                600, // Height
                GraphicsMode.Default,
                "Assessed Coursework",
                GameWindowFlags.Default,
                DisplayDevice.Default,
                3, // major
                3, // minor
                GraphicsContextFlags.ForwardCompatible
                )
        {
        }

        private int[] mVBO_IDs = new int[7];
        private int[] mVAO_IDs = new int[4];
        private ShaderUtility mShader;
        private ModelUtility mBoxUtil, mCylinderUtil, mBallUtil;
        private Matrix4 mView, mEyeView, mGroundModel, mCylinderModel, mBoxModel, mDoomSphereModel, mBallModel;
        DoomSphere objDoomShpere;
        Cylinder objCrossCylinderOne, objCrossCylinderTwo, objCrossTwoCylinderOne, objCrossTwoCylinderTwo, objCylinderOne, objCylinderTwo;
        Box objBoxOne, objBoxTwo, objBoxThree, objBoxFour;
        Balls objBlueBall, objGreenBall;
        private ACWTimer mTimer;
        Vector3 accelerationDueToGravity = new Vector3(0, -0.981f, 0); //-9.81 Normally, but stuff goes REALLY fast like that
        

        protected override void OnLoad(EventArgs e)
        {
            #region Base, Lighting, Camera Stuff, Floor
            mTimer = new ACWTimer();
            mTimer.Start();

            GL.ClearColor(Color4.White);
            GL.Enable(EnableCap.DepthTest);
            GL.Enable(EnableCap.CullFace);

            mShader = new ShaderUtility(@"ACW/Shaders/vPassThrough.vert", @"ACW/Shaders/fLighting.frag");
            GL.UseProgram(mShader.ShaderProgramID);

            mView = Matrix4.CreateTranslation(0, -1.5f, 0);
            int uView = GL.GetUniformLocation(mShader.ShaderProgramID, "uView");
            GL.UniformMatrix4(uView, true, ref mView);

            mEyeView = mView;
            int uEyePosition = GL.GetUniformLocation(mShader.ShaderProgramID, "uEyePosition");
            GL.UniformMatrix4(uEyePosition, true, ref mEyeView);

            int vPositionLocation = GL.GetAttribLocation(mShader.ShaderProgramID, "vPosition");
            int vNormal = GL.GetAttribLocation(mShader.ShaderProgramID, "vNormal");
            int uLightDirectionLocation = GL.GetUniformLocation(mShader.ShaderProgramID, "uLightPosition");

            Vector4 lightPosition = Vector4.Transform(new Vector4(2, 8, 0, 1), mView);  //You need to update the light position
            GL.Uniform4(uLightDirectionLocation, lightPosition);                            //everytime you move the camera

            GL.GenVertexArrays(mVAO_IDs.Length, mVAO_IDs);
            GL.GenBuffers(mVBO_IDs.Length, mVBO_IDs);

            float[] vertices = new float[] {-10, 0, -10,0,1,0,
                                             -10, 0, 10,0,1,0,
                                             10, 0, 10,0,1,0,
                                             10, 0, -10,0,1,0,};

            GL.BindVertexArray(mVAO_IDs[0]);
            GL.BindBuffer(BufferTarget.ArrayBuffer, mVBO_IDs[0]);
            GL.BufferData(BufferTarget.ArrayBuffer, (IntPtr)(vertices.Length * sizeof(float)), vertices, BufferUsageHint.StaticDraw);

            int size;
            GL.GetBufferParameter(BufferTarget.ArrayBuffer, BufferParameterName.BufferSize, out size);
            if (vertices.Length * sizeof(float) != size)
            {
                throw new ApplicationException("Vertex data not loaded onto graphics card correctly");
            }

            GL.EnableVertexAttribArray(vPositionLocation);
            GL.VertexAttribPointer(vPositionLocation, 3, VertexAttribPointerType.Float, false, 6 * sizeof(float), 0);
            GL.EnableVertexAttribArray(vNormal);                                                                            //Floor
            GL.VertexAttribPointer(vNormal, 3, VertexAttribPointerType.Float, true, 6 * sizeof(float), 3 * sizeof(float));  //Normals

            mGroundModel = Matrix4.CreateTranslation(0, 0, -5f);
            
            #endregion

            #region Object Utilities
            mBallUtil = ModelUtility.LoadModel(@"Utility/Models/sphere.bin");
            mCylinderUtil = ModelUtility.LoadModel(@"Utility/Models/cylinder.bin");
            mBoxUtil = ModelUtility.LoadModel(@"Utility/Models/lab22model.sjg");
            #endregion

            #region Buffer Object Calls
            //Balls
            GL.BindVertexArray(mVAO_IDs[1]);
            GL.BindBuffer(BufferTarget.ArrayBuffer, mVBO_IDs[1]);
            GL.BufferData(BufferTarget.ArrayBuffer, (IntPtr)(mBallUtil.Vertices.Length * sizeof(float)), mBallUtil.Vertices, BufferUsageHint.StaticDraw);
            GL.BindBuffer(BufferTarget.ElementArrayBuffer, mVBO_IDs[2]);
            GL.BufferData(BufferTarget.ElementArrayBuffer, (IntPtr)(mBallUtil.Indices.Length * sizeof(float)), mBallUtil.Indices, BufferUsageHint.StaticDraw);

            GL.GetBufferParameter(BufferTarget.ArrayBuffer, BufferParameterName.BufferSize, out size);
            if (mBallUtil.Vertices.Length * sizeof(float) != size)
            {
                throw new ApplicationException("Vertex data not loaded onto graphics card correctly");
            }

            GL.GetBufferParameter(BufferTarget.ElementArrayBuffer, BufferParameterName.BufferSize, out size);
            if (mBallUtil.Indices.Length * sizeof(float) != size)
            {
                throw new ApplicationException("Index data not loaded onto graphics card correctly");
            }

            GL.EnableVertexAttribArray(vPositionLocation);
            GL.VertexAttribPointer(vPositionLocation, 3, VertexAttribPointerType.Float, false, 6 * sizeof(float), 0);
            GL.EnableVertexAttribArray(vNormal);                                                                                
            GL.VertexAttribPointer(vNormal, 3, VertexAttribPointerType.Float, true, 6 * sizeof(float), 3 * sizeof(float));

            //Cylinder
            GL.BindVertexArray(mVAO_IDs[2]);
            GL.BindBuffer(BufferTarget.ArrayBuffer, mVBO_IDs[3]);
            GL.BufferData(BufferTarget.ArrayBuffer, (IntPtr)(mCylinderUtil.Vertices.Length * sizeof(float)), mCylinderUtil.Vertices, BufferUsageHint.StaticDraw);
            GL.BindBuffer(BufferTarget.ElementArrayBuffer, mVBO_IDs[4]);
            GL.BufferData(BufferTarget.ElementArrayBuffer, (IntPtr)(mCylinderUtil.Indices.Length * sizeof(float)), mCylinderUtil.Indices, BufferUsageHint.StaticDraw);

            GL.GetBufferParameter(BufferTarget.ArrayBuffer, BufferParameterName.BufferSize, out size);
            if (mCylinderUtil.Vertices.Length * sizeof(float) != size)
            {
                throw new ApplicationException("Vertex data not loaded onto graphics card correctly");
            }

            GL.GetBufferParameter(BufferTarget.ElementArrayBuffer, BufferParameterName.BufferSize, out size);
            if (mCylinderUtil.Indices.Length * sizeof(float) != size)
            {
                throw new ApplicationException("Index data not loaded onto graphics card correctly");
            }

            GL.EnableVertexAttribArray(vPositionLocation);
            GL.VertexAttribPointer(vPositionLocation, 3, VertexAttribPointerType.Float, false, 6 * sizeof(float), 0);
            GL.EnableVertexAttribArray(vNormal);                                                                                
            GL.VertexAttribPointer(vNormal, 3, VertexAttribPointerType.Float, true, 6 * sizeof(float), 3 * sizeof(float));

            //Boxes
            GL.BindVertexArray(mVAO_IDs[3]);
            GL.BindBuffer(BufferTarget.ArrayBuffer, mVBO_IDs[5]);
            GL.BufferData(BufferTarget.ArrayBuffer, (IntPtr)(mBoxUtil.Vertices.Length * sizeof(float)), mBoxUtil.Vertices, BufferUsageHint.StaticDraw);
            GL.BindBuffer(BufferTarget.ElementArrayBuffer, mVBO_IDs[6]);
            GL.BufferData(BufferTarget.ElementArrayBuffer, (IntPtr)(mBoxUtil.Indices.Length * sizeof(float)), mBoxUtil.Indices, BufferUsageHint.StaticDraw);

            GL.GetBufferParameter(BufferTarget.ArrayBuffer, BufferParameterName.BufferSize, out size);
            if (mBoxUtil.Vertices.Length * sizeof(float) != size)
            {
                throw new ApplicationException("Vertex data not loaded onto graphics card correctly");
            }

            GL.GetBufferParameter(BufferTarget.ElementArrayBuffer, BufferParameterName.BufferSize, out size);
            if (mBoxUtil.Indices.Length * sizeof(float) != size)
            {
                throw new ApplicationException("Index data not loaded onto graphics card correctly");
            }

            GL.EnableVertexAttribArray(vPositionLocation);
            GL.VertexAttribPointer(vPositionLocation, 3, VertexAttribPointerType.Float, false, 6 * sizeof(float), 0);
            GL.EnableVertexAttribArray(vNormal);                                                                               
            GL.VertexAttribPointer(vNormal, 3, VertexAttribPointerType.Float, true, 6 * sizeof(float), 3 * sizeof(float));
            #endregion

            #region Object Setup  
            objBoxOne = new Box(new Vector3(0, 7, 0), true, ref mBoxUtil, 1, 0, ref mBoxModel);
            objBoxTwo = new Box(new Vector3(0, 5, 0), false, ref mBoxUtil, 1, 0, ref mBoxModel);
            objBoxThree = new Box(new Vector3(0, 3, 0), false, ref mBoxUtil, 1, 0, ref mBoxModel);
            objBoxFour = new Box(new Vector3(0, 1, 0), false, ref mBoxUtil, 1, 0, ref mBoxModel);
            objBoxOne.OnLoad(e, ref mVAO_IDs, ref mVBO_IDs);
            objBoxTwo.OnLoad(e, ref mVAO_IDs, ref mVBO_IDs);
            objBoxThree.OnLoad(e, ref mVAO_IDs, ref mVBO_IDs);
            objBoxFour.OnLoad(e, ref mVAO_IDs, ref mVBO_IDs);
            objCrossCylinderOne = new Cylinder(new Vector3(0.075f, 1, 0.075f), new Vector3(0, 5.5f, 0), 1.57f, 0f, mCylinderUtil, ref mCylinderModel);
            objCrossCylinderTwo = new Cylinder(new Vector3(0.075f, 1f, 0.075f), new Vector3(0, 5.5f, 0), 1.57f, 1.5f, mCylinderUtil, ref mCylinderModel);
            objCrossTwoCylinderOne = new Cylinder(new Vector3(0.15f, 1, 0.15f), new Vector3(0, 4.5f, 0), 1.57f, 0f, mCylinderUtil, ref mCylinderModel);
            objCrossTwoCylinderTwo = new Cylinder(new Vector3(0.15f, 1, 0.15f), new Vector3(0, 4.5f, 0), 1.57f, 1.5f, mCylinderUtil, ref mCylinderModel);
            objCylinderOne = new Cylinder(new Vector3(0.15f, 1, 0.15f), new Vector3(0, 3.5f, 0), 1.57f, 1, mCylinderUtil, ref mCylinderModel);
            objCylinderTwo = new Cylinder(new Vector3(0.15f, 1, 0.15f), new Vector3(0, 2.5f, 0), 1.57f, 2, mCylinderUtil, ref mCylinderModel);
            objCrossCylinderOne.OnLoad(e);
            objCrossCylinderTwo.OnLoad(e);
            objCrossTwoCylinderOne.OnLoad(e);
            objCrossTwoCylinderTwo.OnLoad(e);
            objCylinderOne.OnLoad(e);
            objCylinderTwo.OnLoad(e);
            objDoomShpere = new DoomSphere(new Vector3(0, 1, 0), Color4.Yellow, 0.6f, mBallUtil, ref mDoomSphereModel);
            objBlueBall = new Balls(ref mBallUtil, 0.0014f, new Vector3(RandomVel(), RandomVel(), RandomVel()), 0.08f, 0f, new Vector3(-0.3f, 7f, 0f), Color4.DodgerBlue, ref mBallModel);
            objGreenBall = new Balls(ref mBallUtil, 0.001f, new Vector3(RandomVel(), RandomVel(), RandomVel()), 0.04f, 0f, new Vector3(0.3f, 7f, 0f), Color4.ForestGreen, ref mBallModel);
            
            
            #endregion

            base.OnLoad(e);
        }

        private int RandomVel()
        {
            Random rand = new Random();
            int number = rand.Next(-6, 6);

            return number;
        }

        private float RandomPos()
        {
            Random rand = new Random();
            int number = rand.Next(-10, 10);

            float res = number / 10;

            return res;
        }

        protected override void OnKeyPress(KeyPressEventArgs e)
        {
            base.OnKeyPress(e);
            if (e.KeyChar == 'w')
            {
                mView = mView * Matrix4.CreateTranslation(0.0f, 0.0f, 0.1f);
                OnKeyPressData();
            }
            if (e.KeyChar == 's')
            {
                mView = mView * Matrix4.CreateTranslation(0.0f, 0.0f, -0.1f);
                OnKeyPressData();
            }
            if (e.KeyChar == 'a')
            {
                mView = mView * Matrix4.CreateRotationY(-0.03f);
                OnKeyPressData();
            }
            if (e.KeyChar == 'd')
            {
                mView = mView * Matrix4.CreateRotationY(0.03f);
                OnKeyPressData();
            }
            if (e.KeyChar == 'z')
            {
                mView = mView * Matrix4.CreateTranslation(0.0f, -0.1f, 0.0f);
                OnKeyPressData();
            }
            if (e.KeyChar == 'x')
            {
                mView = mView * Matrix4.CreateTranslation(0.0f, 0.1f, 0.0f);
                OnKeyPressData();
            }
                objBlueBall.OnKeyPress(e);
                objGreenBall.OnKeyPress(e);
        }

        private void OnKeyPressData()
        {
            int uLightDirectionLocation = GL.GetUniformLocation(mShader.ShaderProgramID, "uLightPosition");
            Vector4 lightPosition = Vector4.Transform(new Vector4(2, 8, 0, 1), mView);
            GL.Uniform4(uLightDirectionLocation, lightPosition);
            int uView = GL.GetUniformLocation(mShader.ShaderProgramID, "uView");
            GL.UniformMatrix4(uView, true, ref mView);
            int uEyePosition = GL.GetUniformLocation(mShader.ShaderProgramID, "uEyePosition");
            GL.UniformMatrix4(uEyePosition, true, ref mView);
        }

        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);
            GL.Viewport(this.ClientRectangle);
            if (mShader != null)
            {
                int uProjectionLocation = GL.GetUniformLocation(mShader.ShaderProgramID, "uProjection");
                Matrix4 projection = Matrix4.CreatePerspectiveFieldOfView(1, (float)ClientRectangle.Width / ClientRectangle.Height, 0.5f, 25);
                GL.UniformMatrix4(uProjectionLocation, true, ref projection);
            }
        }

        protected override void OnUpdateFrame(FrameEventArgs e)
        {
 	        base.OnUpdateFrame(e);

            #region Data to Pass to
            float timestep = mTimer.GetElapsedSeconds();

            Vector3 greenBallPos = objGreenBall.position; float greenBallRadius = objGreenBall.radius;
            Vector3 blueBallPos = objBlueBall.position; float blueBallRadius = objBlueBall.radius;
            Vector3 doomSpherePos = objDoomShpere.position; float doomSphereRadius = objDoomShpere.radius;

            Vector3 cylinderCross1P1 = new Vector3(-1, 5.5f, 1);    Vector3 cylinderCross1P2 = new Vector3(1, 5.5f, -1);
            Vector3 cylinderCross2P1 = new Vector3(-1, 5.5f, -1);   Vector3 cylinderCross2P2 = new Vector3(1, 5.5f, 1);
            Vector3 cylinderCross3P1 = new Vector3(0, 4.5f, 1);     Vector3 cylinderCross3P2 = new Vector3(0, 4.5f, -1);
            Vector3 cylinderCross4P1 = new Vector3(-1, 4.5f, 0);    Vector3 cylinderCross4P2 = new Vector3(1, 4.5f, 0);
            Vector3 cylinder1P1 = new Vector3(-1, 3.5f, 1);         Vector3 cylinder1P2 = new Vector3(1, 3.5f, -1);
            Vector3 cylinder2P1 = new Vector3(-1, 2.5f, -1);        Vector3 cylinder2P2 = new Vector3(1, 2.5f, 1);
            #endregion

            //Call the update methods on the moving objects
            objBlueBall.OnUpdateFrame(e, timestep, accelerationDueToGravity, mGroundModel, greenBallPos, greenBallRadius, doomSpherePos, doomSphereRadius, objBoxOne, objBoxTwo, objBoxThree, objBoxFour, cylinderCross1P1, cylinderCross1P2, cylinderCross2P1, cylinderCross2P2, cylinderCross3P1, cylinderCross3P2, cylinderCross4P1, cylinderCross4P2, cylinder1P1, cylinder1P2, cylinder2P1, cylinder2P2, 0.075f, 0.15f);
            objGreenBall.OnUpdateFrame(e, timestep, accelerationDueToGravity, mGroundModel, blueBallPos, blueBallRadius, doomSpherePos, doomSphereRadius, objBoxOne, objBoxTwo, objBoxThree, objBoxFour, cylinderCross1P1, cylinderCross1P2, cylinderCross2P1, cylinderCross2P2, cylinderCross3P1, cylinderCross3P2, cylinderCross4P1, cylinderCross4P2, cylinder1P1, cylinder1P2, cylinder2P1, cylinder2P2, 0.075f, 0.15f);
        }

        protected override void OnRenderFrame(FrameEventArgs e)
        {
            //Call the render methods in objects
            base.OnRenderFrame(e);

            #region Primer & Ground
            GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
            
            int uModel = GL.GetUniformLocation(mShader.ShaderProgramID, "uModel");
            int uColour = GL.GetUniformLocation(mShader.ShaderProgramID, "uModel");
            GL.UniformMatrix4(uModel, true, ref mGroundModel);

            GL.BindVertexArray(mVAO_IDs[0]);
            GL.DrawArrays(PrimitiveType.TriangleFan, 0, 4);
            #endregion

            #region Object Draw Calls
            objBoxOne.OnRenderFrame(e, ref mVAO_IDs, ref mGroundModel, ref uModel, ref mShader);
            objBoxTwo.OnRenderFrame(e, ref mVAO_IDs, ref mGroundModel, ref uModel, ref mShader);
            objBoxThree.OnRenderFrame(e, ref mVAO_IDs, ref mGroundModel, ref uModel, ref mShader);
            objBoxFour.OnRenderFrame(e, ref mVAO_IDs, ref mGroundModel, ref uModel, ref mShader);
            objCrossCylinderOne.OnRenderFrame(e, ref mGroundModel, ref uModel, ref mShader, ref mVAO_IDs);
            objCrossCylinderTwo.OnRenderFrame(e, ref mGroundModel, ref uModel, ref mShader, ref mVAO_IDs);
            objCrossTwoCylinderOne.OnRenderFrame(e, ref mGroundModel, ref uModel, ref mShader, ref mVAO_IDs);
            objCrossTwoCylinderTwo.OnRenderFrame(e, ref mGroundModel, ref uModel, ref mShader, ref mVAO_IDs);
            objCylinderOne.OnRenderFrame(e, ref mGroundModel, ref uModel, ref mShader, ref mVAO_IDs);
            objCylinderTwo.OnRenderFrame(e, ref mGroundModel, ref uModel, ref mShader, ref mVAO_IDs);            
            objDoomShpere.OnRenderFrame(e,ref uModel,ref mShader,ref mVAO_IDs, ref uColour);
            objBlueBall.OnRenderFrame(e, ref mGroundModel, ref uModel, ref mShader, ref mVAO_IDs, ref uColour);
            objGreenBall.OnRenderFrame(e, ref mGroundModel, ref uModel, ref mShader, ref mVAO_IDs, ref uColour);
            
            #endregion

            GL.BindVertexArray(0);
            this.SwapBuffers();
        }

        protected override void OnUnload(EventArgs e)
        {
            GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
            GL.BindBuffer(BufferTarget.ElementArrayBuffer, 0);
            GL.BindVertexArray(0);
            GL.DeleteBuffers(mVBO_IDs.Length, mVBO_IDs);
            GL.DeleteVertexArrays(mVAO_IDs.Length, mVAO_IDs);
            mShader.Delete();
            base.OnUnload(e);
        }
    }
}

//To Do:
/*==
 *Create the box object (Manipulate floor probably)         === (Nicked from lab 2, cheers Simon)
 *Setup all the methods in each object                      === (Done)
 *Create a list of all objects                              === (Done)
 *Position all items correctly, sizes and such              === (Done)
 *Set up a state system to reset the positions of the balls === (Done)
 *Collisions                                                === (Done)
    >>Basic Velocity                                        === (Done)
    >>Gravity                                               === (Done)
    >>Collisions on Box                                     === (Done, fixed bug)
    >>Collisions on Cylinder                                === (Done?)
    >>Collisions on DoomSphere                              === (Done, Basic)
    >>Collisions on EachOther                               === (Done, not tested)
    >>Random Velcoity from start                            === (Done)
 *Lighting                                                  ===(Done)<<This much to pass>>
 *Shaders                                                   ===
 *Textures                                                  ===
  ==*/