﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Labs.Utility;
using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;

namespace Labs.ACW
{
    class Balls : NonStaticObject
    {
        public Vector3 position, velocity, initalPos, initalVel;
        public Color4 color;
        public float mass, rotation, radius, initalRadius;
        public float shinyness;
        private ModelUtility ballModelUtil;
        public Matrix4 mBallModelMat;
        private int number;

        public Balls(ref ModelUtility mBallUtil, float inMass, Vector3 inVelocity, float inRadius, float inRotation, Vector3 inPosition, Color4 inColor, ref Matrix4 inModel)
        {
            position = inPosition;
            radius = inRadius;
            initalRadius = inRadius;
            rotation = inRotation;
            ballModelUtil = mBallUtil;         
            velocity = inVelocity;
            color = inColor;
            mass = inMass;
            initalPos = inPosition;
            initalVel = inVelocity;
            mBallModelMat = inModel;
        }

        public virtual void OnLoad(EventArgs e, ref int[] mVAO_IDs, ref int[] mVBO_IDs)
        {
            //Bind buffers
            //Set position
            //Colour
            //Radius
            //Shinyness
            //Mass            
        }

        public virtual void OnKeyPress(KeyPressEventArgs e)
        {
            if (e.KeyChar == 'r')
            {
                position = initalPos;
                velocity = initalVel;
                radius = initalRadius;
            }
        }

        public virtual void OnUpdateFrame(FrameEventArgs e, float timestep, Vector3 accelGravity, Matrix4 mGroundModel, 
            Vector3 otherBallPos, float otherBallRadius, 
            Vector3 doomSpherePos, float doomSphereRadius, 
            Box box1, Box box2, Box box3, Box box4, 
            Vector3 cylinderCross1P1, Vector3 cylinderCross1P2, Vector3 cylinderCross2P1, Vector3 cylinderCross2P2, 
            Vector3 cylinderCross3P1, Vector3 cylinderCross3P2, Vector3 cylinderCross4P1, Vector3 cylinderCross4P2, 
            Vector3 cylinder1P1, Vector3 cylinder1P2, Vector3 cylinder2P1, Vector3 cylinder2P2, float cylinder1Radius, float cylinder2Radius)
        {
            Vector3 oldPos = position;

            velocity = velocity + accelGravity * timestep;
            position = position + velocity * timestep;

            double distance = Math.Sqrt(Math.Pow(position.X - otherBallPos.X, 2) + Math.Pow(otherBallPos.Y - position.Y, 2));
            double distanceToDoom = Math.Sqrt(Math.Pow(position.X - doomSpherePos.X, 2) + Math.Pow(doomSpherePos.Y - position.Y, 2));
            if (distance < (radius + otherBallRadius))
            {
                position = oldPos;
                Vector3 normalP = (position - otherBallPos).Normalized();
                velocity = velocity - 2 * Vector3.Dot(normalP, velocity) * normalP;
            }
            if (distanceToDoom < (radius + doomSphereRadius))
            {
                position = oldPos;
                //Vector3 normalP = (position - doomSpherePos).Normalized();
                //velocity = velocity - 2 * Vector3.Dot(normalP, velocity) * normalP;
                radius = radius / 2;
            }

            float thing = position.Y + (radius / mGroundModel.ExtractScale().Y);

            if (thing < -1)
            {
                position = initalPos;
                radius = initalRadius;
            }

            boxCollideWithRoof(box1.mBoxModel, ref oldPos);
            boxCollide(box2.mBoxModel, ref oldPos);
            boxCollide(box3.mBoxModel, ref oldPos);
            boxCollide(box4.mBoxModel, ref oldPos);
            CylinderCollide(ref oldPos, cylinder1Radius, cylinderCross1P1, cylinderCross1P2);
            CylinderCollide(ref oldPos, cylinder1Radius, cylinderCross2P1, cylinderCross2P2);
            CylinderCollide(ref oldPos, cylinder1Radius, cylinderCross3P1, cylinderCross3P2);
            CylinderCollide(ref oldPos, cylinder1Radius, cylinderCross4P1, cylinderCross4P2);
            CylinderCollide(ref oldPos, cylinder1Radius, cylinder1P1, cylinder1P2);
            CylinderCollide(ref oldPos, cylinder1Radius, cylinder2P1, cylinder2P2);
        }

        #region Box Collide
        private void boxCollide(Matrix4 box, ref Vector3 oldPos)
        {
            Vector3 ballInBoxSpace = Vector3.Transform(position, box.Inverted());

            if (ballInBoxSpace.X + (radius / box.ExtractScale().X) > 1)
            {
                position = oldPos;
                Vector3 normal = Vector3.Transform(new Vector3(-1, 0, 0), box.ExtractRotation());
                velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
            }
            if (ballInBoxSpace.X + (radius / box.ExtractScale().X) < -1)
            {
                position = oldPos;
                Vector3 normal = Vector3.Transform(new Vector3(1, 0, 0), box.ExtractRotation());
                velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
            }
            if (ballInBoxSpace.Z + (radius / box.ExtractScale().Z) > 1)
            {
                position = oldPos;
                Vector3 normal = Vector3.Transform(new Vector3(0, 0, -1), box.ExtractRotation());
                velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
            }
            if (ballInBoxSpace.Z + (radius / box.ExtractScale().Z) < -1)
            {
                position = oldPos;
                Vector3 normal = Vector3.Transform(new Vector3(0, 0, 1), box.ExtractRotation());
                velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
            }
        }
        private void boxCollideWithRoof(Matrix4 box, ref Vector3 oldPos)
        {
            Vector3 ballInBoxSpace = Vector3.Transform(position, box.Inverted());

            if (ballInBoxSpace.X + (radius / box.ExtractScale().X) > 1)
            {
                position = oldPos;
                Vector3 normal = Vector3.Transform(new Vector3(-1, 0, 0), box.ExtractRotation());
                velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
            }
            if (ballInBoxSpace.X + (radius / box.ExtractScale().X) < -1)
            {
                position = oldPos;
                Vector3 normal = Vector3.Transform(new Vector3(1, 0, 0), box.ExtractRotation());
                velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
            }
            if (ballInBoxSpace.Z + (radius / box.ExtractScale().Z) > 1)
            {
                position = oldPos;
                Vector3 normal = Vector3.Transform(new Vector3(0, 0, -1), box.ExtractRotation());
                velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
            }
            if (ballInBoxSpace.Z + (radius / box.ExtractScale().Z) < -1)
            {
                position = oldPos;
                Vector3 normal = Vector3.Transform(new Vector3(0, 0, 1), box.ExtractRotation());
                velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
            }
            if (ballInBoxSpace.Y + (radius / box.ExtractScale().Y) > 1)
            {
                position = oldPos;
                Vector3 normal = Vector3.Transform(new Vector3(0, -1, 0), box.ExtractRotation());
                velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
            }
        }
        #endregion

        private void CylinderCollide(ref Vector3 oldPos, float cylinderRadius, Vector3 pointOne, Vector3 pointTwo)
        {
            Vector3 A = Vector3.Dot((position - pointTwo), (pointOne - pointTwo).Normalized()) * (pointOne - pointTwo).Normalized();

            Vector3 sub = Vector3.Subtract(pointOne, pointTwo);

            if (A.Y > 0 && A.Y < sub.Length)
            {
                Vector3 dist = pointTwo + A - position;
                if (dist.Length < (radius + cylinderRadius))
                {
                    Vector3 normal = (position - dist).Normalized();
                    velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
                }
            }
            if (A.X > 0 && A.X < sub.Length)
            {
                Vector3 dist = pointTwo + A - position;
                if (dist.Length < (radius + cylinderRadius))
                {
                    Vector3 normal = (position - dist).Normalized();
                    velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
                }
            }
            if (A.Z > 0 && A.Z < sub.Length)
            {
                Vector3 dist = pointTwo + A - position;
                if (dist.Length < (radius + cylinderRadius))
                {
                    Vector3 normal = (position - dist).Normalized();
                    velocity = velocity - 2 * Vector3.Dot(normal, velocity) * normal;
                }
            }
        }

        public virtual void OnRenderFrame(FrameEventArgs e, ref Matrix4 mGroundModel, ref int uModel, ref ShaderUtility mShader, ref int[] mVAO_IDs, ref int uColour)
        {
            GL.Uniform4(uColour, color);
            mBallModelMat = Matrix4.CreateScale(radius) * Matrix4.CreateTranslation(position);

            //Matrix4 m = mBallModelMat * mGroundModel;
            uModel = GL.GetUniformLocation(mShader.ShaderProgramID, "uModel");
            GL.UniformMatrix4(uModel, true, ref mBallModelMat);

            GL.BindVertexArray(mVAO_IDs[1]);
            GL.DrawElements(PrimitiveType.Triangles, ballModelUtil.Indices.Length, DrawElementsType.UnsignedInt, 0);
        }

    }
}

//NOTES FOR THIS OBJECT
//>Ball 1===
    // >>Radius = 4
    // >>Density = 0.0014kg
    // >>Colour = ForestGreen
    // >>Shiny = 30
//>Ball 2===
    // >>Radius = 8
    // >>Density = 0.001kg
    // >>Colour = DodgerBlue
    // >>Shiny = 0