﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Labs.Utility;
using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;

namespace Labs.ACW
{
    class Cylinder : StaticObject
    {
        public Vector3 position;
        public float rotationX, rotationY;
        public Vector3 scale;

        private ModelUtility cylModelUtil;
        public Matrix4 cylModel;

        /// <summary>
        /// Constructor for Cylinder object
        /// </summary>
        /// <param name="inPos"></param>
        /// <param name="inRadius"></param>
        /// <param name="inRotate"></param>
        /// <param name="cylType">1 for Cross 1, 2 for Cross 2</param>
        public Cylinder(Vector3 inRadius, Vector3 inPosition, float inRotateX, float inRotateY, ModelUtility cylUtil, ref Matrix4 inCylModel)
        {            
            scale = inRadius;
            rotationX = inRotateX;
            rotationY = inRotateY;
            position = inPosition;
            cylModelUtil = cylUtil;
            cylModel = inCylModel;
        }

        public virtual void OnLoad(EventArgs e)
        {
            //Bind buffers
            //Set position
            //Colour
            //Radius



            cylModel = Matrix4.CreateScale(scale) * Matrix4.CreateRotationX(rotationX) * Matrix4.CreateRotationY(rotationY) * Matrix4.CreateTranslation(position);
        }

        public virtual void OnRenderFrame(FrameEventArgs e, ref Matrix4 mGroundModel, ref int uModel, ref ShaderUtility mShader, ref int[] mVAO_IDs)
        {
            Matrix4 cylinder = Matrix4.CreateScale(scale) * Matrix4.CreateRotationX(rotationX) * Matrix4.CreateRotationY(rotationY) * Matrix4.CreateTranslation(position);
            uModel = GL.GetUniformLocation(mShader.ShaderProgramID, "uModel");
            GL.UniformMatrix4(uModel, true, ref cylinder);

            GL.BindVertexArray(mVAO_IDs[2]);
            GL.DrawElements(PrimitiveType.Triangles, cylModelUtil.Indices.Length, DrawElementsType.UnsignedInt, 0);
        }

    }
}

//NOTES FOR THIS OBJECT
//>CROSS 1===
    // >>Radius = 7.5
    // >>Position = 25% up
    // >>Rotation
//>CROSS 2===
    // >>Radius = 15
    // >>Position = 75% below
    // >>Rotation
//>CYLINDER 1===
    // >>Radius = 15
    // >>Position = 25cm up
    // >>Rotation = 45 Degrees
//>CYLINDER 2===
    // >>Radius = 15
    // >>Position = 25cm below
    // >>Rotation = 0 Degrees