﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Labs.Utility;
using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;

namespace Labs.ACW
{
    class DoomSphere : StaticObject
    {
        public Vector3 position;
        private Color4 color;
        public float radius;
        private ModelUtility sphereModelUtil;
        public Matrix4 ballMatrix;
        public DoomSphere(Vector3 inPos, Color4 inColor, float inRadius, ModelUtility doomShpereModelUtil, ref Matrix4 inBallMatrix)
        {
            position = inPos;
            color = inColor;
            radius = inRadius;
            sphereModelUtil = doomShpereModelUtil;
            ballMatrix = inBallMatrix;
        }

        public virtual void OnLoad(EventArgs e)
        {
            //Bind buffers
            //Set position
            //Colour
            //Radius
            ballMatrix = Matrix4.CreateScale(radius) * Matrix4.CreateTranslation(position);
        }

        public virtual void OnRenderFrame(FrameEventArgs e, ref int uModel, ref ShaderUtility mShader, ref int[] mVAO_IDs, ref int uColour)
        {
            GL.Uniform4(uColour, color);
            ballMatrix = Matrix4.CreateScale(radius) * Matrix4.CreateTranslation(position);
                        
            uModel = GL.GetUniformLocation(mShader.ShaderProgramID, "uModel");
            GL.UniformMatrix4(uModel, true, ref ballMatrix);

            GL.BindVertexArray(mVAO_IDs[1]);
            GL.DrawElements(PrimitiveType.Triangles, sphereModelUtil.Indices.Length, DrawElementsType.UnsignedInt, 0);
        }
    }
}

//NOTES FOR THIS OBJECT
//>Radius = 30
//>Colour = Yellow