﻿using System;
using Labs.Utility;
using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;

namespace Labs.Lab4
{
    public class Lab4_1Window : GameWindow
    {
        private int[] mVertexArrayObjectIDArray = new int[2];
        private int[] mVertexBufferObjectIDArray = new int[2];
        private ShaderUtility mShader;
        private Matrix4 mSquareMatrix,mSquareTwoMatrix;
        private Vector3 mCirclePosition, mCircleVelocity, mCirTwoPosition;
        private float mCircleRadius, mCirTwoRadius;
        private Timer mTimer;

        public Lab4_1Window()
            : base(
                800, // Width
                600, // Height
                GraphicsMode.Default,
                "Lab 4_1 Simple Animation and Collision Detection",
                GameWindowFlags.Default,
                DisplayDevice.Default,
                3, // major
                3, // minor
                GraphicsContextFlags.ForwardCompatible
                )
        {
        }

        protected override void OnLoad(EventArgs e)
        {
            GL.ClearColor(Color4.AliceBlue);

            mTimer = new Timer();
            mTimer.Start();

            mCircleRadius = 0.1f;
            mCirTwoRadius = 0.2f;

            mShader = new ShaderUtility(@"Lab4/Shaders/vLab4.vert", @"Lab4/Shaders/fLab4.frag");
            int vPositionLocation = GL.GetAttribLocation(mShader.ShaderProgramID, "vPosition");
            GL.UseProgram(mShader.ShaderProgramID);

            float[] vertices = new float[] { 
                   -1f, -1f,    //Left Down
                   1f, -1f,     //Right Down
                   1f, 1f,      //Right Up
                   -1f, 1f      //Left Up
            };

            GL.GenVertexArrays(mVertexArrayObjectIDArray.Length, mVertexArrayObjectIDArray);
            GL.GenBuffers(mVertexBufferObjectIDArray.Length, mVertexBufferObjectIDArray);

            GL.BindVertexArray(mVertexArrayObjectIDArray[0]);
            GL.BindBuffer(BufferTarget.ArrayBuffer, mVertexBufferObjectIDArray[0]);
            GL.BufferData(BufferTarget.ArrayBuffer, (IntPtr)(vertices.Length * sizeof(float)), vertices, BufferUsageHint.StaticDraw);

            int size;
            GL.GetBufferParameter(BufferTarget.ArrayBuffer, BufferParameterName.BufferSize, out size);

            if (vertices.Length * sizeof(float) != size)
            {
                throw new ApplicationException("Vertex data not loaded onto graphics card correctly");
            }

            GL.EnableVertexAttribArray(vPositionLocation);
            GL.VertexAttribPointer(vPositionLocation, 2, VertexAttribPointerType.Float, false, 2 * sizeof(float), 0);

            vertices = new float[200];

            for (int i = 0; i < 100; ++i)
            {
                vertices[2 * i] = (float)Math.Cos(MathHelper.DegreesToRadians(i * 360.0 / 100));
                vertices[2 * i + 1] = (float)Math.Cos(MathHelper.DegreesToRadians(90.0 + i * 360.0 / 100));
            }

            GL.BindVertexArray(mVertexArrayObjectIDArray[1]);
            GL.BindBuffer(BufferTarget.ArrayBuffer, mVertexBufferObjectIDArray[1]);

            GL.BufferData(BufferTarget.ArrayBuffer, (IntPtr)(vertices.Length * sizeof(float)), vertices, BufferUsageHint.StaticDraw);

            GL.GetBufferParameter(BufferTarget.ArrayBuffer, BufferParameterName.BufferSize, out size);

            if (vertices.Length * sizeof(float) != size)
            {
                throw new ApplicationException("Vertex data not loaded onto graphics card correctly");
            }

            GL.EnableVertexAttribArray(vPositionLocation);
            GL.VertexAttribPointer(vPositionLocation, 2, VertexAttribPointerType.Float, false, 2 * sizeof(float), 0);

            int uViewLocation = GL.GetUniformLocation(mShader.ShaderProgramID, "uView");
            Matrix4 m = Matrix4.CreateTranslation(0, 0, 0);
            GL.UniformMatrix4(uViewLocation, true, ref m);

            mSquareMatrix = Matrix4.CreateScale(4,4,4) * Matrix4.CreateRotationZ(0f) * Matrix4.CreateTranslation(0f, 0f, 0);
            mSquareTwoMatrix = Matrix4.CreateScale(1f) * Matrix4.CreateRotationZ(0f) * Matrix4.CreateTranslation(0f, 0, 0);



            mCirclePosition = new Vector3(1.5f, 1, 0);

            Random rand = new Random();
            int speed = rand.Next(-6, 6);
            int speedTwo = rand.Next(-6, 6);

            mCircleVelocity = new Vector3(speed, speedTwo, 0);
            

            mCirTwoPosition = new Vector3(2f, 1f, 0);

            base.OnLoad(e);
        }

        private void SetCamera()
        {
            float height = ClientRectangle.Height;
            float width = ClientRectangle.Width;
            if (mShader != null)
            {
                Matrix4 proj;
                if (height > width)
                {
                    if (width == 0)
                    {
                        width = 1;
                    }
                    proj = Matrix4.CreateOrthographic(10, 10 * height / width, 0, 10);
                }
                else
                {
                    if (height == 0)
                    {
                        height = 1;
                    }
                    proj = Matrix4.CreateOrthographic(10 * width / height, 10, 0, 10);
                }
                int uProjectionLocation = GL.GetUniformLocation(mShader.ShaderProgramID, "uProjection");
                GL.UniformMatrix4(uProjectionLocation, true, ref proj);
            }
        }

        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);
            GL.Viewport(this.ClientRectangle);
            SetCamera();
        }

        protected override void OnRenderFrame(FrameEventArgs e)
        {
            base.OnRenderFrame(e);
            GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

            int uModelMatrixLocation = GL.GetUniformLocation(mShader.ShaderProgramID, "uModel");
            int uColourLocation = GL.GetUniformLocation(mShader.ShaderProgramID, "uColour");

            //Blue Sqaure
            GL.Uniform4(uColourLocation, Color4.DodgerBlue);

            GL.UniformMatrix4(uModelMatrixLocation, true, ref mSquareMatrix);
            GL.BindVertexArray(mVertexArrayObjectIDArray[0]);
            GL.DrawArrays(PrimitiveType.LineLoop, 0, 4);

            GL.UniformMatrix4(uModelMatrixLocation, true, ref mSquareTwoMatrix);
            GL.BindVertexArray(mVertexArrayObjectIDArray[0]);
            GL.DrawArrays(PrimitiveType.LineLoop, 0, 4);

            //Circle One
            Matrix4 circleMatrix = Matrix4.CreateScale(mCircleRadius) * Matrix4.CreateTranslation(mCirclePosition);

            GL.UniformMatrix4(uModelMatrixLocation, true, ref circleMatrix);
            GL.BindVertexArray(mVertexArrayObjectIDArray[1]);
            GL.DrawArrays(PrimitiveType.LineLoop, 0, 100);

            //Circle Two
            Matrix4 circleTwoMatrix = Matrix4.CreateScale(mCirTwoRadius) * Matrix4.CreateTranslation(mCirTwoPosition);

            GL.UniformMatrix4(uModelMatrixLocation, true, ref circleTwoMatrix);
            GL.BindVertexArray(mVertexArrayObjectIDArray[1]);
            GL.DrawArrays(PrimitiveType.LineLoop, 0, 100);

            //Red Sqaure
            GL.Uniform4(uColourLocation, Color4.Red);

            Matrix4 m = mSquareMatrix * mSquareMatrix.Inverted();

            GL.UniformMatrix4(uModelMatrixLocation, true, ref m);
            GL.BindVertexArray(mVertexArrayObjectIDArray[0]);
            GL.DrawArrays(PrimitiveType.LineLoop, 0, 4);

            m = mSquareTwoMatrix * mSquareMatrix.Inverted();

            GL.UniformMatrix4(uModelMatrixLocation, true, ref m);
            GL.BindVertexArray(mVertexArrayObjectIDArray[0]);
            GL.DrawArrays(PrimitiveType.LineLoop, 0, 4);

            //Moving Circle
            m = (Matrix4.CreateScale(mCircleRadius) * Matrix4.CreateTranslation(mCirclePosition)) * mSquareMatrix.Inverted();

            GL.UniformMatrix4(uModelMatrixLocation, true, ref m); 
            GL.BindVertexArray(mVertexArrayObjectIDArray[1]); 
            GL.DrawArrays(PrimitiveType.LineLoop, 0, 100);

            //Second Circle
            m = (Matrix4.CreateScale(mCirTwoRadius) * Matrix4.CreateTranslation(mCirTwoPosition)) * mSquareMatrix.Inverted();

            GL.UniformMatrix4(uModelMatrixLocation, true, ref m);
            GL.BindVertexArray(mVertexArrayObjectIDArray[1]);
            GL.DrawArrays(PrimitiveType.LineLoop, 0, 100);

            this.SwapBuffers();
        }

        protected override void OnUpdateFrame(FrameEventArgs e) 
        { 
            base.OnUpdateFrame(e);
            Vector3 tempCirPos;
            float timestep = mTimer.GetElapsedSeconds();
            
            //Set a temp position inside the square
            tempCirPos = mCirclePosition;

            mCirclePosition = mCirclePosition + mCircleVelocity * timestep;
            Vector3 circleInSquareSpace = Vector3.Transform(mCirclePosition, mSquareMatrix.Inverted());

            //verlet

            //NextPos = 2*NowPos - OldPos + accel * timestep;

            double distance =  Math.Sqrt(Math.Pow(mCirclePosition.X - mCirTwoPosition.X, 2) + Math.Pow(mCirTwoPosition.Y - mCirclePosition.Y, 2));


            //Collision            
            if (distance < (mCircleRadius + mCirTwoRadius))
            {
                mCirclePosition = tempCirPos;
                Vector3 normalP = (mCirclePosition - mCirTwoPosition).Normalized();
                mCircleVelocity = mCircleVelocity - 2 * Vector3.Dot(normalP, mCircleVelocity) * normalP;
            }
            if (circleInSquareSpace.X + (mCircleRadius / mSquareMatrix.ExtractScale().X) > 1) 
            {
                mCirclePosition = tempCirPos;
                //mCircleVelocity.X = -mCircleVelocity.X;
                Vector3 normal = Vector3.Transform(new Vector3(-1, 0, 0), mSquareMatrix.ExtractRotation()); 
                mCircleVelocity = mCircleVelocity - 2 * Vector3.Dot(normal, mCircleVelocity) * normal;
            }
            if (circleInSquareSpace.X - (mCircleRadius / mSquareMatrix.ExtractScale().X) < -1)
            {
                mCirclePosition = tempCirPos;
                //mCircleVelocity.X = -mCircleVelocity.X;
                Vector3 normal = Vector3.Transform(new Vector3(1, 0, 0), mSquareMatrix.ExtractRotation());
                mCircleVelocity = mCircleVelocity - 2 * Vector3.Dot(normal, mCircleVelocity) * normal;
            }
            if (circleInSquareSpace.Y + (mCircleRadius / mSquareMatrix.ExtractScale().Y) > 1)
            {
                mCirclePosition = tempCirPos;
                //mCircleVelocity.Y = -mCircleVelocity.Y;
                Vector3 normal = Vector3.Transform(new Vector3(0, -1, 0), mSquareMatrix.ExtractRotation());
                mCircleVelocity = mCircleVelocity - 2 * Vector3.Dot(normal, mCircleVelocity) * normal;
            }
            if (circleInSquareSpace.Y - (mCircleRadius / mSquareMatrix.ExtractScale().Y) < -1)
            {
                mCirclePosition = tempCirPos;
                //mCircleVelocity.Y = -mCircleVelocity.Y;
                Vector3 normal = Vector3.Transform(new Vector3(0, 1, 0), mSquareMatrix.ExtractRotation());
                mCircleVelocity = mCircleVelocity - 2 * Vector3.Dot(normal, mCircleVelocity) * normal;
            }

            //THIS IS WHERE L41T19 IS (KINDA FIXED)
            //The maths are right, but it stops on the surrounding sqaure's inner edge
            //when the if statement is "dist.Length < mCircleRadius".
            //Other bugs include stopping in the middle of nowhere.

            Vector3 leftDown = new Vector3(-1, -1, -0);
            Vector3 leftUp = new Vector3(-1, 1, 0);
            Vector3 rightDown = new Vector3(1f, -1f,0);
            Vector3 rightUP = new Vector3(1f, 1f,0);

            InnerSqaureSideEdge(ref tempCirPos, ref circleInSquareSpace, leftUp, leftDown);         //Left Edge
            InnerSqaureTopBotEdge(ref tempCirPos, ref circleInSquareSpace, rightDown, leftDown);    //Bottom Edge
            InnerSqaureSideEdge(ref tempCirPos, ref circleInSquareSpace, rightUP, rightDown);     //Right Edge
            InnerSqaureTopBotEdge(ref tempCirPos, ref circleInSquareSpace, rightUP, leftUp);        //Top Edge
            
        }

        private void InnerSqaureSideEdge(ref Vector3 tempCirPos, ref Vector3 circleInSquareSpace, Vector3 pointOne, Vector3 pointTwo)
        {
            Vector3 A = Vector3.Dot((mCirclePosition - pointTwo), (pointOne - pointTwo).Normalized()) * (pointOne - pointTwo).Normalized();

            Vector3 sub = Vector3.Subtract(pointOne, pointTwo);

            if (A.Y > 0 && A.Y < sub.Length)
            {
                Vector3 dist = pointTwo + A - mCirclePosition;
                if (dist.Length < mCircleRadius)
                {
                    //mCirclePosition = tempCirPos;
                    Vector3 normal = (mCirclePosition - dist).Normalized();
                    mCircleVelocity = mCircleVelocity - 2 * Vector3.Dot(normal, mCircleVelocity) * normal;
                }
            }

            VertCollide(ref tempCirPos, ref pointOne, ref pointTwo);
        }

        private void VertCollide(ref Vector3 tempCirPos, ref Vector3 pointOne, ref Vector3 pointTwo)
        {
            double distance = Math.Sqrt(Math.Pow(mCirclePosition.X - pointOne.X, 2) + Math.Pow(pointOne.Y - mCirclePosition.Y, 2));
            double distanceTwo = Math.Sqrt(Math.Pow(mCirclePosition.X - pointTwo.X, 2) + Math.Pow(pointTwo.Y - mCirclePosition.Y, 2));

            if (distance < (mCircleRadius + 0))
            {
                //mCirclePosition = tempCirPos;
                Vector3 normal = (mCirclePosition - pointOne).Normalized();
                mCircleVelocity = mCircleVelocity - 2 * Vector3.Dot(normal, mCircleVelocity) * normal;
            }
            else if (distanceTwo < (mCircleRadius + 0))
            {
                //mCirclePosition = tempCirPos;
                Vector3 normal = (mCirclePosition - pointTwo).Normalized();
                mCircleVelocity = mCircleVelocity - 2 * Vector3.Dot(normal, mCircleVelocity) * normal;
            }
        }

        private void InnerSqaureTopBotEdge(ref Vector3 tempCirPos, ref Vector3 circleInSquareSpace, Vector3 pointOne, Vector3 pointTwo)
        {
            Vector3 A = Vector3.Dot((mCirclePosition - pointTwo), (pointOne - pointTwo).Normalized()) * (pointOne - pointTwo).Normalized();

            Vector3 sub = Vector3.Subtract(pointOne, pointTwo);

            if (A.X > 0 && A.X < sub.Length)
            {
                Vector3 dist = pointTwo + A - mCirclePosition;
                if (dist.Length < mCircleRadius)
                {
                    //mCirclePosition = tempCirPos;
                    Vector3 normal = (mCirclePosition - dist).Normalized();
                    mCircleVelocity = mCircleVelocity - 2 * Vector3.Dot(normal, mCircleVelocity) * normal;
                }
            }
            VertCollide(ref tempCirPos, ref pointOne, ref pointTwo);
        }

        protected override void OnUnload(EventArgs e)
        {
            base.OnUnload(e);
            GL.DeleteBuffers(mVertexBufferObjectIDArray.Length, mVertexBufferObjectIDArray);
            GL.DeleteVertexArrays(mVertexArrayObjectIDArray.Length, mVertexArrayObjectIDArray);
            GL.UseProgram(0);
            mShader.Delete();
        }
    }
}